@@ -1,8 +1,5 @@
 /* fhandler.cc.  See console.cc for fhandler_console functions.
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
-   2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015 Red Hat, Inc.
-
 This file is part of Cygwin.
 
 This software is a copyrighted work licensed under the terms of the
@@ -13,7 +10,7 @@ details. */
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/uio.h>
-#include <sys/acl.h>
+#include <cygwin/acl.h>
 #include <sys/param.h>
 #include "cygerrno.h"
 #include "perprocess.h"
@@ -162,8 +159,8 @@ char *fhandler_base::get_proc_fd_name (char *buf)
 {
   if (get_name ())
     return strcpy (buf, get_name ());
-  if (dev ().name)
-    return strcpy (buf, dev ().name);
+  if (dev ().name ())
+    return strcpy (buf, dev ().name ());
   return strcpy (buf, "");
 }
 
