@@ -25,7 +25,6 @@ details. */
 #include "cygtls.h"
 #include "tls_pbuf.h"
 #include "child_info.h"
-#include "dll_init.h"
 
 class pinfo_basic: public _pinfo
 {
@@ -35,7 +34,7 @@ public:
 
 pinfo_basic::pinfo_basic ()
 {
-  dwProcessId = GetCurrentProcessId ();
+  pid = dwProcessId = GetCurrentProcessId ();
   PWCHAR pend = wcpncpy (progname, global_progname,
 			 sizeof (progname) / sizeof (WCHAR) - 1);
   *pend = L'\0';
@@ -55,41 +54,42 @@ void
 pinfo::thisproc (HANDLE h)
 {
   procinfo = NULL;
-  bool execed = !!h;
 
   DWORD flags = PID_IN_USE | PID_ACTIVE;
-  /* Forked process or process started from non-Cygwin parent needs a pid. */
-  if (!execed)
+  if (!h)
     {
-      cygheap->pid = create_cygwin_pid ();
-      flags |= PID_NEW;
       h = INVALID_HANDLE_VALUE;
+      cygheap->pid = cygwin_pid (myself_initial.pid);
+      flags |= PID_NEW;
     }
-  /* spawnve'd process got pid in parent, cygheap->pid has been set in
-     child_info_spawn::handle_spawn. */
 
   init (cygheap->pid, flags, h);
-  InterlockedOr ((LONG *) &procinfo->process_state, PID_IN_USE);
-  procinfo->dwProcessId = myself_initial.dwProcessId;
+  procinfo->process_state |= PID_IN_USE;
+  procinfo->dwProcessId = myself_initial.pid;
   procinfo->sendsig = myself_initial.sendsig;
   wcscpy (procinfo->progname, myself_initial.progname);
-  if (!execed)
-    create_winpid_symlink ();
-  procinfo->exec_sendsig = NULL;
-  procinfo->exec_dwProcessId = 0;
   debug_printf ("myself dwProcessId %u", procinfo->dwProcessId);
+  if (h != INVALID_HANDLE_VALUE)
+    {
+      /* here if execed */
+      static pinfo NO_COPY myself_identity;
+      myself_identity.init (cygwin_pid (procinfo->dwProcessId), PID_EXECED, NULL);
+      procinfo->exec_sendsig = NULL;
+      procinfo->exec_dwProcessId = 0;
+      myself_identity->ppid = procinfo->pid;
+    }
 }
 
 /* Initialize the process table entry for the current task.
    This is not called for forked tasks, only execed ones.  */
-void
+void __stdcall
 pinfo_init (char **envp, int envc)
 {
   if (envp)
     {
       environ_init (envp, envc);
       /* spawn has already set up a pid structure for us so we'll use that */
-      InterlockedOr ((LONG *) &myself->process_state, PID_CYGPARENT);
+      myself->process_state |= PID_CYGPARENT;
     }
   else
     {
@@ -97,42 +97,36 @@ pinfo_init (char **envp, int envc)
 
       myself.thisproc (NULL);
       myself->pgid = myself->sid = myself->pid;
-      myself->ctty = CTTY_UNINITIALIZED;
+      myself->ctty = -1;
       myself->uid = ILLEGAL_UID;
       myself->gid = ILLEGAL_GID;
       environ_init (NULL, 0);	/* call after myself has been set up */
       myself->nice = winprio_to_nice (GetPriorityClass (GetCurrentProcess ()));
-      myself->sched_policy = SCHED_OTHER;
-      myself->sched_reset_on_fork = false;
       myself->ppid = 1;		/* always set last */
       debug_printf ("Set nice to %d", myself->nice);
     }
 
-  InterlockedOr ((LONG *) &myself->process_state, PID_ACTIVE);
-  InterlockedAnd ((LONG *) &myself->process_state,
-		  ~(PID_INITIALIZING | PID_EXITED | PID_REAPED));
-  if (being_debugged ())
-    InterlockedOr ((LONG *) &myself->process_state, PID_DEBUGGED);
+  myself->process_state |= PID_ACTIVE;
+  myself->process_state &= ~(PID_INITIALIZING | PID_EXITED | PID_REAPED);
   myself.preserve ();
-  debug_printf ("pid %d, pgid %d, process_state %y",
-		myself->pid, myself->pgid, myself->process_state);
+  debug_printf ("pid %d, pgid %d, process_state %y", myself->pid, myself->pgid, myself->process_state);
 }
 
 DWORD
 pinfo::status_exit (DWORD x)
 {
-  switch ((NTSTATUS) x)
+  switch (x)
     {
     case STATUS_DLL_NOT_FOUND:
       {
 	path_conv pc;
 	if (!procinfo)
-	   pc.check ("/dev/null", PC_POSIX);
+	   pc.check ("/dev/null", PC_NOWARN | PC_POSIX);
 	else
 	  {
 	    UNICODE_STRING uc;
 	    RtlInitUnicodeString(&uc, procinfo->progname);
-	    pc.check (&uc, PC_POSIX);
+	    pc.check (&uc, PC_NOWARN | PC_POSIX);
 	  }
 	small_printf ("%s: error while loading shared libraries: %s: cannot "
 		      "open shared object file: No such file or directory\n",
@@ -158,14 +152,11 @@ pinfo::status_exit (DWORD x)
 	 reason (but note, the environment *in* CMD is broken and shortened).
 	 This occurs at a point where there's no return to the exec'ing parent
 	 process, so we have to find some way to inform the user what happened.
-
+	 
 	 FIXME: For now, just return with SIGBUS set.  Maybe it's better to add
 	 a lengthy small_printf instead. */
       x = SIGBUS;
       break;
-    case STATUS_CONTROL_C_EXIT:
-      x = SIGINT;
-      break;
     default:
       debug_printf ("*** STATUS_%y\n", x);
       x = 127 << 8;
@@ -215,7 +206,7 @@ pinfo::exit (DWORD n)
     maybe_set_exit_code_from_windows ();	/* may block */
   exit_state = ES_FINAL;
 
-  if (CTTY_IS_VALID (myself->ctty) && !iscons_dev (myself->ctty))
+  if (myself->ctty > 0 && !iscons_dev (myself->ctty))
     {
       lock_ttys here;
       tty *t = cygwin_shared->tty[device::minor(myself->ctty)];
@@ -231,8 +222,6 @@ pinfo::exit (DWORD n)
   int exitcode = self->exitcode & 0xffff;
   if (!self->cygstarted)
     exitcode = ((exitcode & 0xff) << 8) | ((exitcode >> 8) & 0xff);
-  sigproc_printf ("Calling dlls.cleanup_forkables n %y, exitcode %y", n, exitcode);
-  dlls.cleanup_forkables ();
   sigproc_printf ("Calling ExitProcess n %y, exitcode %y", n, exitcode);
   if (!TerminateProcess (GetCurrentProcess (), exitcode))
     system_printf ("TerminateProcess failed, %E");
@@ -240,106 +229,6 @@ pinfo::exit (DWORD n)
 }
 # undef self
 
-/* Return next free Cygwin PID between 2 and MAX_PID, round-robin.  Each new
-   PID is checked that it doesn't collide with an existing PID.  For that,
-   just check if the "cygpid.PID" section exists. */
-pid_t
-create_cygwin_pid ()
-{
-  pid_t pid = 0;
-  WCHAR sym_name[24];
-  UNICODE_STRING sym_str;
-  OBJECT_ATTRIBUTES attr;
-  HANDLE sym_hdl;
-  NTSTATUS status;
-
-  do
-    {
-      do
-	{
-	  pid = ((uint32_t) InterlockedIncrement (&cygwin_shared->pid_src))
-		% MAX_PID;
-	}
-      while (pid < 2);
-      __small_swprintf (sym_name, L"cygpid.%u", pid);
-      RtlInitUnicodeString (&sym_str, sym_name);
-      InitializeObjectAttributes (&attr, &sym_str, OBJ_CASE_INSENSITIVE,
-				  get_shared_parent_dir (), NULL);
-      /* We just want to know if the section (and thus the process) still
-         exists.  Instead of actually opening the section, try to open
-	 it as symlink.  NtOpenSymbolicLinkObject will always returns an
-	 error:
-	 - STATUS_OBJECT_NAME_NOT_FOUND if the section doesn't exist,
-	   so the slot is free and we can use this pid.
-	 - STATUS_OBJECT_TYPE_MISMATCH if the section exists, so we have
-	   to skip this pid and loop to try the next one.
-	  As side-effect we never have to close the section handle and thus
-	  we don't influence the lifetime of the section. */
-      status = NtOpenSymbolicLinkObject (&sym_hdl, SYMBOLIC_LINK_QUERY, &attr);
-    }
-  while (status == STATUS_OBJECT_TYPE_MISMATCH);
-  return pid;
-}
-
-/* Convert Windows WINPID into Cygwin PID.  Utilize the "winpid.WINPID"
-   symlinks created for each process.  The symlink contains the Cygwin
-   PID as target.  Return 0 if no "winpid.WINPID" symlink exists for
-   this WINPID. */
-pid_t
-cygwin_pid (DWORD dwProcessId)
-{
-  WCHAR sym_name[24];
-  WCHAR pid_name[12];
-  UNICODE_STRING sym_str;
-  UNICODE_STRING pid_str;
-  OBJECT_ATTRIBUTES attr;
-  HANDLE sym_hdl;
-  NTSTATUS status;
-
-  __small_swprintf (sym_name, L"winpid.%u", dwProcessId);
-  RtlInitUnicodeString (&sym_str, sym_name);
-  InitializeObjectAttributes (&attr, &sym_str, OBJ_CASE_INSENSITIVE,
-			      get_shared_parent_dir (), NULL);
-  status = NtOpenSymbolicLinkObject (&sym_hdl, SYMBOLIC_LINK_QUERY, &attr);
-  if (!NT_SUCCESS (status))
-    return 0;
-  RtlInitEmptyUnicodeString (&pid_str, pid_name,
-			     sizeof pid_name - sizeof (WCHAR));
-  status = NtQuerySymbolicLinkObject (sym_hdl, &pid_str, NULL);
-  NtClose (sym_hdl);
-  if (!NT_SUCCESS (status))
-    {
-      system_printf ("NtOpenSymbolicLinkObject: %y, PID %u, ret 0",
-		     status, dwProcessId);
-      return 0;
-    }
-  pid_str.Buffer[pid_str.Length / sizeof (WCHAR)] = L'\0';
-  pid_t ret = (pid_t) wcstoul (pid_str.Buffer, NULL, 10);
-  return ret;
-}
-
-/* Create "winpid.WINPID" symlinks with the Cygwin PID of that process as
-   target.  This is used to find the Cygwin PID for a given Windows WINPID. */
-void
-pinfo::create_winpid_symlink ()
-{
-  WCHAR sym_name[24];
-  WCHAR pid_name[24];
-  UNICODE_STRING sym_str;
-  UNICODE_STRING pid_str;
-  OBJECT_ATTRIBUTES attr;
-
-  __small_swprintf (sym_name, L"winpid.%u", procinfo->dwProcessId);
-  RtlInitUnicodeString (&sym_str, sym_name);
-  __small_swprintf (pid_name, L"%u", procinfo->pid);
-  RtlInitUnicodeString (&pid_str, pid_name);
-  InitializeObjectAttributes (&attr, &sym_str, OBJ_CASE_INSENSITIVE,
-			      get_shared_parent_dir (),
-			      everyone_sd (SYMBOLIC_LINK_QUERY));
-  NtCreateSymbolicLinkObject (&winpid_hdl, SYMBOLIC_LINK_ALL_ACCESS,
-			      &attr, &pid_str);
-}
-
 inline void
 pinfo::_pinfo_release ()
 {
@@ -363,18 +252,20 @@ pinfo::init (pid_t n, DWORD flag, HANDLE h0)
 {
   shared_locations shloc;
   h = NULL;
-  if (myself && n == myself->pid)
+  if (myself && !(flag & PID_EXECED)
+      && (n == myself->pid || (DWORD) n == myself->dwProcessId))
     {
       procinfo = myself;
       destroy = 0;
       return;
     }
 
-  int createit = (flag & PID_IN_USE);
+  int createit = flag & (PID_IN_USE | PID_EXECED);
   DWORD access = FILE_MAP_READ
-		 | (flag & (PID_IN_USE | PID_MAP_RW) ? FILE_MAP_WRITE : 0);
+		 | (flag & (PID_IN_USE | PID_EXECED | PID_MAP_RW)
+		    ? FILE_MAP_WRITE : 0);
   if (!h0 || myself.h)
-    shloc = (flag & PID_IN_USE) ? SH_JUSTCREATE : SH_JUSTOPEN;
+    shloc = (flag & (PID_IN_USE | PID_EXECED)) ? SH_JUSTCREATE : SH_JUSTOPEN;
   else
     {
       shloc = SH_MYSELF;
@@ -390,10 +281,14 @@ pinfo::init (pid_t n, DWORD flag, HANDLE h0)
 
   for (int i = 0; i < 20; i++)
     {
-      bool created;
+      DWORD mapsize;
+      if (flag & PID_EXECED)
+	mapsize = PINFO_REDIR_SIZE;
+      else
+	mapsize = sizeof (_pinfo);
 
-      procinfo = (_pinfo *) open_shared (L"cygpid", n, h0, sizeof (_pinfo),
-					 shloc, created, sec_attribs, access);
+      procinfo = (_pinfo *) open_shared (L"cygpid", n, h0, mapsize, &shloc,
+					 sec_attribs, access);
       if (!h0)
 	{
 	  if (createit)
@@ -414,10 +309,17 @@ pinfo::init (pid_t n, DWORD flag, HANDLE h0)
 	  continue;
 	}
 
-      /* Just fetching info for ps or /proc, don't do anything rash. */
-      if (!created && !(flag & PID_NEW) && !procinfo->ppid
-	  && (flag & PID_PROCINFO))
-	break;
+      bool created = shloc != SH_JUSTOPEN;
+
+      /* Detect situation where a transitional memory block is being retrieved.
+	 If the block has been allocated with PINFO_REDIR_SIZE but not yet
+	 updated with a PID_EXECED state then we'll retry.  */
+      if (!created && !(flag & PID_NEW))
+	/* If not populated, wait 2 seconds for procinfo to become populated.
+	   Would like to wait with finer granularity but that is not easily
+	   doable.  */
+	for (int i = 0; i < 200 && !procinfo->ppid; i++)
+	  Sleep (10);
 
       if (!created && createit && (procinfo->process_state & PID_REAPED))
 	{
@@ -426,18 +328,32 @@ pinfo::init (pid_t n, DWORD flag, HANDLE h0)
 				   shared memory */
 	}
 
-      if (procinfo->process_state & PID_REAPED)
+      if ((procinfo->process_state & PID_REAPED)
+	  || ((procinfo->process_state & PID_INITIALIZING) && (flag & PID_NOREDIR)
+	      && cygwin_pid (procinfo->dwProcessId) != procinfo->pid))
 	{
 	  set_errno (ESRCH);
 	  break;
 	}
 
+      if (procinfo->process_state & PID_EXECED)
+	{
+	  pid_t realpid = procinfo->pid;
+	  debug_printf ("execed process windows pid %u, cygwin pid %d", n, realpid);
+	  if (realpid == n)
+	    api_fatal ("retrieval of execed process info for pid %d failed due to recursion.", n);
+
+	  n = realpid;
+	  CloseHandle (h0);
+	  h0 = NULL;
+	  goto loop;
+	}
+
       /* In certain pathological cases, it is possible for the shared memory
 	 region to exist for a while after a process has exited.  This should
 	 only be a brief occurrence, so rather than introduce some kind of
 	 locking mechanism, just loop.  */
-      if (!created && createit
-	  && (procinfo->process_state & (PID_EXITED | PID_REAPED)))
+      if (!created && createit && (procinfo->process_state & (PID_EXITED | PID_REAPED)))
 	{
 	  debug_printf ("looping because pid %d, procinfo->pid %d, "
 			"procinfo->dwProcessid %u has PID_EXITED|PID_REAPED set",
@@ -447,8 +363,15 @@ pinfo::init (pid_t n, DWORD flag, HANDLE h0)
 
       if (flag & PID_NEW)
 	procinfo->start_time = time (NULL);
-      if (created)
+      if (!created)
+	/* nothing */;
+      else if (!(flag & PID_EXECED))
 	procinfo->pid = n;
+      else
+	{
+	  procinfo->process_state |= PID_IN_USE | PID_EXECED;
+	  procinfo->pid = myself->pid;
+	}
 
       h = h0;	/* Success! */
       break;
@@ -469,8 +392,6 @@ pinfo::init (pid_t n, DWORD flag, HANDLE h0)
       h = h0;
       _pinfo_release ();
     }
-  if (shloc == SH_MYSELF)
-    cygheap->shared_regions.myself_shared_addr = procinfo;
 }
 
 void
@@ -490,35 +411,40 @@ pinfo::set_acl()
     debug_printf ("NtSetSecurityObject %y", status);
 }
 
-void
-pinfo_minimal::set_inheritance (bool inherit)
-{
-  DWORD i_flag = inherit ? HANDLE_FLAG_INHERIT : 0;
-
-  SetHandleInformation (rd_proc_pipe, HANDLE_FLAG_INHERIT, i_flag);
-  SetHandleInformation (hProcess, HANDLE_FLAG_INHERIT, i_flag);
-  SetHandleInformation (h, HANDLE_FLAG_INHERIT, i_flag);
-}
-
 pinfo::pinfo (HANDLE parent, pinfo_minimal& from, pid_t pid):
   pinfo_minimal (), destroy (false), procinfo (NULL), waiter_ready (false),
   wait_thread (NULL)
 {
-  /* cygheap_exec_info::record_children set the inheritance of the required
-     child handles so just copy them over... */
-  rd_proc_pipe = from.rd_proc_pipe;
-  hProcess = from.hProcess;
-  h = from.h;
-  /* ...and reset their inheritance. */
-  set_inheritance (false);
-  init (pid, PID_MAP_RW, h);
+  HANDLE herr;
+  const char *duperr = NULL;
+  if (!DuplicateHandle (parent, herr = from.rd_proc_pipe, GetCurrentProcess (),
+			&rd_proc_pipe, 0, false, DUPLICATE_SAME_ACCESS))
+    duperr = "couldn't duplicate parent rd_proc_pipe handle %p for forked child %d after exec, %E";
+  else if (!DuplicateHandle (parent, herr = from.hProcess, GetCurrentProcess (),
+			     &hProcess, 0, false, DUPLICATE_SAME_ACCESS))
+    duperr = "couldn't duplicate parent process handle %p for forked child %d after exec, %E";
+  else
+    {
+      h = NULL;
+      DuplicateHandle (parent, from.h, GetCurrentProcess (), &h, 0, false,
+		       DUPLICATE_SAME_ACCESS);
+      init (pid, PID_MAP_RW, h);
+      if (*this)
+	return;
+    }
+
+  if (duperr)
+    debug_printf (duperr, herr, pid);
+
+  /* Returning with procinfo == NULL.  Any open handles will be closed by the
+     destructor. */
 }
 
 const char *
 _pinfo::_ctty (char *buf)
 {
-  if (!CTTY_IS_VALID (ctty))
-    stpcpy (buf, "no ctty");
+  if (ctty <= 0)
+    strcpy (buf, "no ctty");
   else
     {
       device d;
@@ -533,32 +459,26 @@ _pinfo::set_ctty (fhandler_termios *fh, int flags)
 {
   tty_min& tc = *fh->tc ();
   debug_printf ("old %s, ctty device number %y, tc.ntty device number %y flags & O_NOCTTY %y", __ctty (), ctty, tc.ntty, flags & O_NOCTTY);
-  if (fh && (!CTTY_IS_VALID (ctty) || ctty == tc.ntty) && !(flags & O_NOCTTY))
+  if (fh && (ctty <= 0 || ctty == tc.ntty) && !(flags & O_NOCTTY))
     {
-      if (tc.getsid () && tc.getsid () != sid && ctty == CTTY_RELEASED)
-	; /* Do nothing if another session is associated with the TTY. */
-      else
+      ctty = tc.ntty;
+      if (cygheap->ctty != fh->archetype)
 	{
-	  ctty = tc.ntty;
-	  if (cygheap->ctty != fh->archetype)
+	  debug_printf ("cygheap->ctty %p, archetype %p", cygheap->ctty, fh->archetype);
+	  if (!cygheap->ctty)
+	    syscall_printf ("ctty was NULL");
+	  else
 	    {
-	      debug_printf ("cygheap->ctty %p, archetype %p",
-			    cygheap->ctty, fh->archetype);
-	      if (!cygheap->ctty)
-		syscall_printf ("ctty was NULL");
-	      else
-		{
-		  syscall_printf ("ctty %p, usecount %d", cygheap->ctty,
-				  cygheap->ctty->archetype_usecount (0));
-		  cygheap->ctty->close ();
-		}
-	      cygheap->ctty = (fhandler_termios *) fh->archetype;
-	      if (cygheap->ctty)
-		{
-		  fh->archetype_usecount (1);
-		  /* guard ctty fh */
-		  report_tty_counts (cygheap->ctty, "ctty", "");
-		}
+	      syscall_printf ("ctty %p, usecount %d", cygheap->ctty,
+			      cygheap->ctty->archetype_usecount (0));
+	      cygheap->ctty->close ();
+	    }
+	  cygheap->ctty = (fhandler_termios *) fh->archetype;
+	  if (cygheap->ctty)
+	    {
+	      fh->archetype_usecount (1);
+	      /* guard ctty fh */
+	      report_tty_counts (cygheap->ctty, "ctty", "");
 	    }
 	}
 
@@ -566,11 +486,7 @@ _pinfo::set_ctty (fhandler_termios *fh, int flags)
       syscall_printf ("attaching %s sid %d, pid %d, pgid %d, tty->pgid %d, tty->sid %d",
 		      __ctty (), sid, pid, pgid, tc.getpgid (), tc.getsid ());
       if (!cygwin_finished_initializing && !myself->cygstarted
-	  && pgid == pid && tc.getpgid () && tc.getsid ()
-	  /* Even GDB starts app via CreateProcess which changes cygstarted.
-	     This results in setting the wrong pgid here, so just skip this
-	     under debugger. */
-	  && !being_debugged ())
+	  && pgid == pid && tc.getpgid () && tc.getsid ())
 	pgid = tc.getpgid ();
 
       /* May actually need to do this:
@@ -581,22 +497,21 @@ _pinfo::set_ctty (fhandler_termios *fh, int flags)
 	 an obvious bug surfaces. */
       if (sid == pid && !tc.getsid ())
 	tc.setsid (sid);
-      if (CTTY_IS_VALID (ctty))
-	sid = tc.getsid ();
+      sid = tc.getsid ();
       /* See above */
-      if ((!tc.getpgid () || being_debugged ()) && pgid == pid)
+      if (!tc.getpgid () && pgid == pid)
 	tc.setpgid (pgid);
     }
-  debug_printf ("cygheap->ctty now %p, archetype %p", cygheap->ctty, fh ? fh->archetype : NULL);
-  return CTTY_IS_VALID (ctty);
+  debug_printf ("cygheap->ctty now %p, archetype %p", cygheap->ctty, fh->archetype);
+  return ctty > 0;
 }
 
 /* Test to determine if a process really exists and is processing signals.
  */
-bool
+bool __reg1
 _pinfo::exists ()
 {
-  return process_state && !(process_state & (PID_EXITED | PID_REAPED));
+  return this && process_state && !(process_state & (PID_EXITED | PID_REAPED | PID_EXECED));
 }
 
 bool
@@ -609,20 +524,7 @@ _pinfo::alive ()
   return !!h;
 }
 
-static commune_result
-commune_process_siginfo ()
-{
-  commune_result res = { 0 };
-
-  res.pnd = sig_send (myself, __SIGPENDINGALL, NULL);
-  res.blk = cygheap->compute_sigblkmask ();
-  for (int sig = 1; sig < NSIG; ++sig)
-    if (global_sigs[sig].sa_handler == SIG_IGN)
-      res.ign |= SIGTOMASK (sig);
-  return res;
-}
-
-DWORD
+DWORD WINAPI
 commune_process (void *arg)
 {
   siginfo_t& si = *((siginfo_t *) arg);
@@ -645,24 +547,30 @@ commune_process (void *arg)
       {
 	sigproc_printf ("processing PICOM_CMDLINE");
 	unsigned n = 0;
-	int argc = __argv0_orig ? 1 : __argc_safe;
-	const char *argv[argc + 1];
+	const char *argv[__argc_safe + 1];
 
-	for (int i = 0; i < argc; i++)
+	for (int i = 0; i < __argc_safe; i++)
 	  {
-	    argv[i] = __argv[i] ?: "";
+	    if (IsBadStringPtr (__argv[i], INT32_MAX))
+	      argv[i] = "";
+	    else
+	      argv[i] = __argv[i];
 	    n += strlen (argv[i]) + 1;
 	  }
-	argv[argc] = NULL;
+	argv[__argc_safe] = NULL;
 	if (!WritePipeOverlapped (tothem, &n, sizeof n, &nr, 1000L))
-	  sigproc_printf ("WritePipeOverlapped sizeof argv failed, %E");
-	else for (int i = 0; i < argc; i++)
-	  if (!WritePipeOverlapped (tothem, __argv[i],
-				    strlen (__argv[i]) + 1, &nr, 1000L))
-	    {
-	      sigproc_printf ("WritePipeOverlapped arg %d failed, %E", i);
-	      break;
-	    }
+	  {
+	    /*__seterrno ();*/	// this is run from the signal thread, so don't set errno
+	    sigproc_printf ("WritePipeOverlapped sizeof argv failed, %E");
+	  }
+	else
+	  for (const char **a = argv; *a; a++)
+	    if (!WritePipeOverlapped (tothem, *a, strlen (*a) + 1, &nr, 1000L))
+	      {
+		sigproc_printf ("WritePipeOverlapped arg %d failed, %E",
+				a - argv);
+		break;
+	      }
 	break;
       }
     case PICOM_CWD:
@@ -689,14 +597,6 @@ commune_process (void *arg)
 	  sigproc_printf ("WritePipeOverlapped root failed, %E");
 	break;
       }
-    case PICOM_SIGINFO:
-      {
-	sigproc_printf ("processing PICOM_SIGINFO");
-	commune_result cr = commune_process_siginfo ();
-	if (!WritePipeOverlapped (tothem, &cr, sizeof cr, &nr, 1000L))
-	  sigproc_printf ("WritePipeOverlapped siginfo failed, %E");
-	break;
-      }
     case PICOM_FDS:
       {
 	sigproc_printf ("processing PICOM_FDS");
@@ -719,7 +619,7 @@ commune_process (void *arg)
       }
     case PICOM_PIPE_FHANDLER:
       {
-	sigproc_printf ("processing PICOM_PIPE_FHANDLER");
+	sigproc_printf ("processing PICOM_FDS");
 	int64_t unique_id = si._si_commune._si_pipe_unique_id;
 	unsigned int n = 0;
 	cygheap_fdenum cfd;
@@ -738,30 +638,6 @@ commune_process (void *arg)
 	  sigproc_printf ("WritePipeOverlapped sizeof hdl failed, %E");
 	break;
       }
-    case PICOM_FILE_PATHCONV:
-      {
-	sigproc_printf ("processing PICOM_FILE_PATHCONV");
-	int fd = si._si_commune._si_fd;
-	uint32_t flags = si._si_commune._si_flags;
-	unsigned int n = 0;
-	cygheap_fdget cfd (fd);
-	if (cfd >= 0
-	    && (!(flags & FFH_LINKAT)
-		|| (cfd->get_flags () & (O_TMPFILE | O_EXCL))
-		    != (O_TMPFILE | O_EXCL)))
-	  {
-	    fhandler_base *fh = cfd;
-	    void *ser_buf = fh->pc.serialize (fh->get_handle (), n);
-	    if (!WritePipeOverlapped (tothem, &n, sizeof n, &nr, 1000L))
-	      sigproc_printf ("WritePipeOverlapped sizeof hdl failed, %E");
-	    else if (!WritePipeOverlapped (tothem, ser_buf, n, &nr, 1000L))
-	      sigproc_printf ("WritePipeOverlapped hdl failed, %E");
-	    cfree (ser_buf);
-	  }
-	else if (!WritePipeOverlapped (tothem, &n, sizeof n, &nr, 1000L))
-	  sigproc_printf ("WritePipeOverlapped sizeof hdl failed, %E");
-	break;
-      }
     case PICOM_FD:
       {
 	sigproc_printf ("processing PICOM_FD");
@@ -778,26 +654,6 @@ commune_process (void *arg)
 	  sigproc_printf ("WritePipeOverlapped fd failed, %E");
 	break;
       }
-    case PICOM_ENVIRON:
-      {
-	sigproc_printf ("processing PICOM_ENVIRON");
-	unsigned n = 0;
-	char **env = environ;
-	if (env)
-	  for (char **e = env; *e; e++)
-	    n += strlen (*e) + 1;
-	if (!WritePipeOverlapped (tothem, &n, sizeof n, &nr, 1000L))
-	  sigproc_printf ("WritePipeOverlapped sizeof argv failed, %E");
-	else if (env)
-	  for (char **e = env; *e; e++)
-	    if (!WritePipeOverlapped (tothem, *e, strlen (*e) + 1, &nr, 1000L))
-	      {
-	        sigproc_printf ("WritePipeOverlapped arg %d failed, %E",
-				e - env);
-	        break;
-	      }
-	break;
-      }
     }
   if (process_sync)
     {
@@ -817,14 +673,17 @@ commune_result
 _pinfo::commune_request (__uint32_t code, ...)
 {
   DWORD nr;
-  commune_result res = { 0 };
+  commune_result res;
   va_list args;
   siginfo_t si = {0};
   HANDLE& hp = si._si_commune._si_process_handle;
   HANDLE& fromthem = si._si_commune._si_read_handle;
   HANDLE request_sync = NULL;
 
-  if (!pid)
+  res.s = NULL;
+  res.n = 0;
+
+  if (!this || !pid)
     {
       set_errno (ESRCH);
       goto err;
@@ -844,9 +703,7 @@ _pinfo::commune_request (__uint32_t code, ...)
       break;
 
     case PICOM_FD:
-    case PICOM_FILE_PATHCONV:
       si._si_commune._si_fd = va_arg (args, int);
-      si._si_commune._si_flags = va_arg (args, uint32_t);
       break;
 
     break;
@@ -873,12 +730,10 @@ _pinfo::commune_request (__uint32_t code, ...)
     {
     case PICOM_CMDLINE:
     case PICOM_CWD:
-    case PICOM_ENVIRON:
     case PICOM_ROOT:
     case PICOM_FDS:
     case PICOM_FD:
     case PICOM_PIPE_FHANDLER:
-    case PICOM_FILE_PATHCONV:
       if (!ReadPipeOverlapped (fromthem, &n, sizeof n, &nr, 1000L)
 	  || nr != sizeof n)
 	{
@@ -903,14 +758,6 @@ _pinfo::commune_request (__uint32_t code, ...)
 	  res.n = p - res.s;
 	}
       break;
-    case PICOM_SIGINFO:
-      if (!ReadPipeOverlapped (fromthem, &res, sizeof res, &nr, 1000L)
-	  || nr != sizeof res)
-	{
-	  __seterrno ();
-	  goto err;
-	}
-      break;
     }
   goto out;
 
@@ -934,7 +781,7 @@ out:
 fhandler_pipe *
 _pinfo::pipe_fhandler (int64_t unique_id, size_t &n)
 {
-  if (!pid)
+  if (!this || !pid)
     return NULL;
   if (pid == myself->pid)
     return NULL;
@@ -943,23 +790,11 @@ _pinfo::pipe_fhandler (int64_t unique_id, size_t &n)
   return (fhandler_pipe *) cr.s;
 }
 
-void *
-_pinfo::file_pathconv (int fd, uint32_t flags, size_t &n)
-{
-  if (!pid)
-    return NULL;
-  if (pid == myself->pid)
-    return NULL;
-  commune_result cr = commune_request (PICOM_FILE_PATHCONV, fd, flags);
-  n = cr.n;
-  return (void *) cr.s;
-}
-
 char *
 _pinfo::fd (int fd, size_t &n)
 {
   char *s;
-  if (!pid)
+  if (!this || !pid)
     return NULL;
   if (pid != myself->pid)
     {
@@ -983,7 +818,7 @@ char *
 _pinfo::fds (size_t &n)
 {
   char *s;
-  if (!pid)
+  if (!this || !pid)
     return NULL;
   if (pid != myself->pid)
     {
@@ -1011,7 +846,7 @@ char *
 _pinfo::root (size_t& n)
 {
   char *s;
-  if (!pid)
+  if (!this || !pid)
     return NULL;
   if (pid != myself->pid && !ISSTATE (this, PID_NOTCYGWIN))
     {
@@ -1030,23 +865,6 @@ _pinfo::root (size_t& n)
   return s;
 }
 
-int
-_pinfo::siginfo (sigset_t &pnd, sigset_t &blk, sigset_t &ign)
-{
-  commune_result cr;
-
-  if (!pid)
-    return -1;
-  if (pid != myself->pid && !ISSTATE (this, PID_NOTCYGWIN))
-    cr = commune_request (PICOM_SIGINFO);
-  else
-    cr = commune_process_siginfo ();
-  pnd = cr.pnd;
-  blk = cr.blk;
-  ign = cr.ign;
-  return -1;
-}
-
 static HANDLE
 open_commune_proc_parms (DWORD pid, PRTL_USER_PROCESS_PARAMETERS prupp)
 {
@@ -1074,7 +892,7 @@ char *
 _pinfo::cwd (size_t& n)
 {
   char *s = NULL;
-  if (!pid)
+  if (!this || !pid)
     return NULL;
   if (ISSTATE (this, PID_NOTCYGWIN))
     {
@@ -1120,7 +938,7 @@ char *
 _pinfo::cmdline (size_t& n)
 {
   char *s = NULL;
-  if (!pid)
+  if (!this || !pid)
     return NULL;
   if (ISSTATE (this, PID_NOTCYGWIN))
     {
@@ -1162,81 +980,25 @@ _pinfo::cmdline (size_t& n)
   else
     {
       n = 0;
-      int argc = __argv0_orig ? 1 : __argc_safe;
-
-      for (int i = 0; i < argc; ++i)
-	n += strlen (__argv[i]) + 1;
+      for (char **a = __argv; *a; a++)
+	n += strlen (*a) + 1;
       char *p;
       p = s = (char *) cmalloc_abort (HEAP_COMMUNE, n);
-      for (int i = 0; i < argc; ++i)
-	p = stpcpy (p, __argv[i]) + 1;
+      for (char **a = __argv; *a; a++)
+	{
+	  strcpy (p, *a);
+	  p = strchr (p, '\0') + 1;
+	}
     }
   return s;
 }
 
-
-char *
-_pinfo::environ (size_t& n)
-{
-  char **env = NULL;
-  if (!pid)
-    return NULL;
-  if (ISSTATE (this, PID_NOTCYGWIN))
-    {
-      RTL_USER_PROCESS_PARAMETERS rupp;
-      HANDLE proc = open_commune_proc_parms (dwProcessId, &rupp);
-
-      if (!proc)
-        return NULL;
-
-      MEMORY_BASIC_INFORMATION mbi;
-      SIZE_T envsize;
-      PWCHAR envblock;
-      if (!VirtualQueryEx (proc, rupp.Environment, &mbi, sizeof(mbi)))
-        {
-          NtClose (proc);
-          return NULL;
-        }
-
-      SIZE_T read;
-      envsize = (ptrdiff_t) mbi.RegionSize
-                - ((ptrdiff_t) rupp.Environment - (ptrdiff_t) mbi.BaseAddress);
-      envblock = (PWCHAR) cmalloc_abort (HEAP_COMMUNE, envsize);
-
-      if (ReadProcessMemory (proc, rupp.Environment, envblock, envsize, &read))
-        env = win32env_to_cygenv (envblock, false);
-
-      NtClose (proc);
-    }
-  else if (pid != myself->pid)
-    {
-      commune_result cr = commune_request (PICOM_ENVIRON);
-      n = cr.n;
-      return cr.s;
-    }
-  else
-    env = ::environ;
-
-  if (env == NULL)
-    return NULL;
-
-  n = 0;
-  for (char **e = env; *e; e++)
-    n += strlen (*e) + 1;
-
-  char *p, *s;
-  p = s = (char *) cmalloc_abort (HEAP_COMMUNE, n);
-  for (char **e = env; *e; e++)
-    p = stpcpy (p, *e) + 1;
-  return s;
-}
-
 /* This is the workhorse which waits for the write end of the pipe
    created during new process creation.  If the pipe is closed or a zero
    is received on the pipe, it is assumed that the cygwin pid has exited.
    Otherwise, various "signals" can be sent to the parent to inform the
    parent to perform a certain action. */
-static DWORD
+static DWORD WINAPI
 proc_waiter (void *arg)
 {
   pinfo vchild = *(pinfo *) arg;
@@ -1379,8 +1141,6 @@ void
 pinfo::release ()
 {
   _pinfo_release ();
-  if (winpid_hdl)
-    NtClose (winpid_hdl);
   HANDLE close_h;
   if (rd_proc_pipe)
     {
@@ -1468,12 +1228,12 @@ winpids::add (DWORD& nelem, bool winpid, DWORD pid)
     {
       npidlist += slop_pidlist;
       pidlist = (DWORD *) realloc (pidlist, size_pidlist (npidlist + 1));
-      pinfolist = (pinfo *) realloc ((void *) pinfolist, size_pinfolist (npidlist + 1));
+      pinfolist = (pinfo *) realloc (pinfolist, size_pinfolist (npidlist + 1));
     }
 
   _onreturn onreturn;
   pinfo& p = pinfolist[nelem];
-  memset ((void *) &p, 0, sizeof (p));
+  memset (&p, 0, sizeof (p));
 
   bool perform_copy;
   if (cygpid == myself->pid)
@@ -1487,16 +1247,12 @@ winpids::add (DWORD& nelem, bool winpid, DWORD pid)
 	 shared memory region. */
       onreturn = OpenProcess (PROCESS_QUERY_LIMITED_INFORMATION, false, pid);
 
-      /* If we couldn't open the process then we don't have rights to it
-	 and should make a copy of the shared memory area when it exists
-	 (it may not).  */
+      /* If we couldn't open the process then we don't have rights to it and should
+	 make a copy of the shared memory area when it exists (it may not).  */
       perform_copy = onreturn ? make_copy : true;
 
-      p.init (cygpid, PID_PROCINFO | pinfo_access, NULL);
+      p.init (cygpid, PID_NOREDIR | pinfo_access, NULL);
     }
-  /* Did we catch the process during exec?  Try to fix. */
-  if (p && p->dwProcessId != pid)
-    pid = p->dwProcessId;
 
   /* If we're just looking for winpids then don't do any special cygwin "stuff* */
   if (winpid)
@@ -1509,40 +1265,47 @@ winpids::add (DWORD& nelem, bool winpid, DWORD pid)
      that it isn't a cygwin process. */
   if (!p)
     {
-      if (!pinfo_access || !cygpid)
+      if (!pinfo_access)
 	return;
-      p.init (cygpid, PID_PROCINFO, NULL);
+      p.init (cygpid, PID_NOREDIR, NULL);
       if (!p)
 	return;
     }
 
-out:
   /* Scan list of previously recorded pids to make sure that this pid hasn't
      shown up before.  This can happen when a process execs. */
   for (unsigned i = 0; i < nelem; i++)
-    if (pidlist[i] == pid)
+    if (pinfolist[i]->pid == p->pid)
       {
-	if (p && (_pinfo *) p != (_pinfo *) myself)
+	if ((_pinfo *) p != (_pinfo *) myself)
 	  p.release ();
 	return;
       }
-  /* If p is "false" then, eventually any opened process handle will be closed
-     and the function will exit without adding anything to the pid list.
+
+out:
+  /* Exit here.
+
+     If p is "false" then, eventually any opened process handle will be closed and
+     the function will exit without adding anything to the pid list.
 
      If p is "true" then we've discovered a cygwin process.
 
      Handle "myself" differently.  Don't copy it and close/zero the handle we
-     just opened to it.  If not performing a copy, then keep the process handle
-     open for the duration of the life of the procinfo region to potential
-     races when a new process uses this pid.  Otherwise, malloc some memory
-     for a copy of the shared memory.
+     just opened to it.
+     If not performing a copy, then keep the process handle open for the duration
+     of the life of the procinfo region to potential races when a new process uses
+     this pid.
+     Otherwise, malloc some memory for a copy of the shared memory.
 
-     If malloc failed, then "oh well".  Just keep the shared memory around
+     If the malloc failed, then "oh well".  Just keep the shared memory around
      and eventually close the handle when the winpids goes out of scope.
 
      If malloc succeeds, copy the procinfo we just grabbed into the new region,
      release the shared memory and allow the handle to be closed when this
-     function returns. */
+     function returns.
+
+     Oh, and add the pid to the list and bump the number of elements.  */
+
   if (p)
     {
       if (p == (_pinfo *) myself)
@@ -1566,9 +1329,8 @@ out:
 	    }
 	}
     }
-  /* Add pid to the list and bump the number of elements.  */
   if (p || winpid)
-    pidlist[nelem++] = pid;
+    pidlist[nelem++] = !p ? pid : p->dwProcessId;
 }
 
 DWORD
@@ -1578,36 +1340,24 @@ winpids::enum_processes (bool winpid)
 
   if (!winpid)
     {
-      tmp_pathbuf tp;
-      NTSTATUS status;
       HANDLE dir = get_shared_parent_dir ();
       BOOLEAN restart = TRUE;
-      bool last_run = false;
-      ULONG context = 0;
-      PDIRECTORY_BASIC_INFORMATION dbi_buf = (PDIRECTORY_BASIC_INFORMATION)
-					     tp.w_get ();
-      while (!last_run)
+      ULONG context;
+      struct fdbi
+	{
+	  DIRECTORY_BASIC_INFORMATION dbi;
+	  WCHAR buf[2][NAME_MAX + 1];
+	} f;
+      while (NT_SUCCESS (NtQueryDirectoryObject (dir, &f, sizeof f, TRUE,
+						 restart, &context, NULL)))
 	{
-	  status = NtQueryDirectoryObject (dir, dbi_buf, 65536, FALSE, restart,
-					   &context, NULL);
-	  if (!NT_SUCCESS (status))
-	    {
-	      debug_printf ("NtQueryDirectoryObject, status %y", status);
-	      break;
-	    }
-	  if (status != STATUS_MORE_ENTRIES)
-	    last_run = true;
 	  restart = FALSE;
-	  for (PDIRECTORY_BASIC_INFORMATION dbi = dbi_buf;
-	       dbi->ObjectName.Length > 0;
-	       dbi++)
+	  f.dbi.ObjectName.Buffer[f.dbi.ObjectName.Length / sizeof (WCHAR)] = L'\0';
+	  if (wcsncmp (f.dbi.ObjectName.Buffer, L"cygpid.", 7) == 0)
 	    {
-	      if (wcsncmp (dbi->ObjectName.Buffer, L"winpid.", 7) == 0)
-		{
-		  DWORD pid = wcstoul (dbi->ObjectName.Buffer + 7, NULL, 10);
-		  add (nelem, false, pid);
-		}
-	    }
+	    DWORD pid = wcstoul (f.dbi.ObjectName.Buffer + 7, NULL, 10);
+	    add (nelem, false, pid);
+	  }
 	}
     }
   else
