@@ -21,6 +21,8 @@ details. */
 #include <wctype.h>
 #include <errno.h>
 
+#define _WIN32_WINNT 0x0a00
+#define WINVER 0x0a00
 #define NOCOMATTRIBUTE
 #include <windows.h>
 #include <userenv.h>
@@ -73,7 +75,7 @@ static struct option long_options[] = {
 
 static char options[] = "ac:df:hilmMopst:uUVwAC:DHOPSWF:";
 
-static void __attribute__ ((__noreturn__))
+static void
 usage (FILE * stream, int status)
 {
   if (!ignore_flag || !status)
@@ -412,7 +414,6 @@ get_short_paths (char *path)
       exit (1);
     }
   my_wcstombs (ptr, sbuf, len);
-  free (sbuf);
   return ptr;
 }
 
@@ -968,6 +969,7 @@ do_options (int argc, char **argv, int from_file)
 
 	case 'h':
 	  usage (stdout, 0);
+	  break;
 
 	case 'V':
 	  print_version ();
