@@ -13,6 +13,7 @@ details. */
 #include <wchar.h>
 #include <getopt.h>
 #include <locale.h>
+#define WINVER 0x0502
 #include <windows.h>
 #include <sys/cygwin.h>
 #include <cygwin/version.h>
@@ -88,7 +89,7 @@ char **argv;
 HKEY key;
 wchar_t *value;
 
-static void __attribute__ ((__noreturn__))
+static void
 usage (FILE *where = stderr)
 {
   fprintf (where, ""
@@ -165,13 +166,11 @@ usage (FILE *where = stderr)
       "  machine  HKLM  HKEY_LOCAL_MACHINE\n"
       "  users    HKU   HKEY_USERS\n"
       "\n"
-      "You can use forward slash ('/') as a separator instead of backslash, in\n"
-      "that case backslash is treated as an escape character.\n"
-      "You can also supply the registry path prefix /proc/registry{,32,64}/ to\n"
-      "use path completion.\n");
+      "If the keyname starts with a forward slash ('/'), the forward slash is used\n"
+      "as separator and the backslash can be used as escape character.\n");
       fprintf (where, ""
       "Example:\n"
-      "%s list '/HKLM/SOFTWARE/Classes/MIME/Database/Content Type/audio\\/wav'\n\n", prog_name);
+      "%s list '/machine/SOFTWARE/Classes/MIME/Database/Content Type/audio\\/wav'\n\n", prog_name);
     }
   if (where == stderr)
     fprintf (where,
@@ -198,14 +197,13 @@ print_version ()
 void
 Fail (unsigned int rv)
 {
-  wchar_t *buf;
-
+  char *buf;
   if (!quiet)
     {
-      FormatMessageW (FORMAT_MESSAGE_ALLOCATE_BUFFER
-		      | FORMAT_MESSAGE_FROM_SYSTEM,
-		      0, rv, 0, (WCHAR *)& buf, 0, 0);
-      fprintf (stderr, "Error (%d): %ls\n", rv, buf);
+      FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER
+		     | FORMAT_MESSAGE_FROM_SYSTEM,
+		     0, rv, 0, (CHAR *) & buf, 0, 0);
+      fprintf (stderr, "Error (%d): %s\n", rv, buf);
       LocalFree (buf);
     }
   exit (1);
@@ -351,15 +349,6 @@ find_key (int howmanyparts, REGSAM access, int option = 0)
       *h = 0;
       n = e;
     }
-  else if (strncmp ("\\proc\\registry", n, strlen ("\\proc\\registry")) == 0)
-    {
-      /* skip /proc/registry{,32,64}/ prefix */
-      n += strlen ("\\proc\\registry");
-      if (strncmp ("64", n, strlen ("64")) == 0)
-        n += strlen ("64");
-      else if (strncmp ("32", n, strlen ("32")) == 0)
-        n += strlen ("32");
-    }
   while (*n != '\\')
     n++;
   *n++ = 0;
@@ -834,7 +823,10 @@ int
 cmd_load ()
 {
   if (!argv[1])
-    usage ();
+    {
+      usage ();
+      return 1;
+    }
   find_key (1, 0);
   return 0;
 }
@@ -843,7 +835,10 @@ int
 cmd_unload ()
 {
   if (argv[1])
-    usage ();
+    {
+      usage ();
+      return 1;
+    }
   find_key (1, 0);
   return 0;
 }
@@ -852,7 +847,10 @@ int
 cmd_save ()
 {
   if (!argv[1])
-    usage ();
+    {
+      usage ();
+      return 1;
+    }
   /* REG_OPTION_BACKUP_RESTORE is necessary to save /HKLM/SECURITY */
   find_key (1, KEY_QUERY_VALUE, REG_OPTION_BACKUP_RESTORE);
   ssize_t len = cygwin_conv_path (CCP_POSIX_TO_WIN_W, argv[1], NULL, 0);
@@ -870,7 +868,10 @@ int
 cmd_restore ()
 {
   if (!argv[1])
-    usage ();
+    {
+      usage ();
+      return 1;
+    }
   /* REG_OPTION_BACKUP_RESTORE is necessary to restore /HKLM/SECURITY */
   find_key (1, KEY_ALL_ACCESS, REG_OPTION_BACKUP_RESTORE);
   ssize_t len = cygwin_conv_path (CCP_POSIX_TO_WIN_W, argv[1], NULL, 0);
@@ -996,4 +997,6 @@ main (int argc, char **_argv)
 	return commands[i].func ();
       }
   usage ();
+
+  return 0;
 }
